""" Generator functions related to any input and output of accumulation kernels

Namely:
* Coefficient containers
* accumulation object (r, jac...)
"""

from dune.perftool.options import get_option
from dune.perftool.generation import (domain,
                                      function_mangler,
                                      iname,
                                      pymbolic_expr,
                                      globalarg,
                                      valuearg,
                                      get_global_context_value
                                      )
from dune.perftool.pdelab import (name_index,
                                  restricted_name,
                                  )
from dune.perftool.pdelab.basis import (evaluate_coefficient,
                                        evaluate_coefficient_gradient,
                                        name_basis,
                                        )
from dune.perftool.pdelab.spaces import (lfs_iname,
                                         name_lfs_bound,
                                         )
from dune.perftool import Restriction

from pymbolic.primitives import Call, Subscript, Variable

from loopy import CallMangleInfo
from loopy.symbolic import FunctionIdentifier
from loopy.types import NumpyType

import numpy


class CoefficientAccess(FunctionIdentifier):
    def __init__(self, container):
        self.container = container

    def __getinitargs__(self):
        return (self.container,)

    @property
    def name(self):
        return self.container


@function_mangler
def coefficient_mangler(target, func, dtypes):
    if isinstance(func, CoefficientAccess):
        return CallMangleInfo(func.name, (NumpyType(numpy.float64),), (NumpyType(str), NumpyType(numpy.int32)))


class PDELabAccumulationFunction(FunctionIdentifier):
    def __init__(self, accumobj, rank):
        self.accumobj = accumobj
        self.rank = rank

        assert rank in (1, 2)

    def __getinitargs__(self):
        return (self.accumobj, self.rank)

    @property
    def name(self):
        return '{}.accumulate'.format(self.accumobj)


@function_mangler
def accumulation_mangler(target, func, dtypes):
    if isinstance(func, PDELabAccumulationFunction):
        if func.rank == 1:
            return CallMangleInfo(func.name,
                                  (),
                                  (NumpyType(str),
                                   NumpyType(numpy.int32),
                                   NumpyType(numpy.float64),
                                   )
                                  )
        if func.rank == 2:
            return CallMangleInfo(func.name,
                                  (),
                                  (NumpyType(str),
                                   NumpyType(numpy.int32),
                                   NumpyType(str),
                                   NumpyType(numpy.int32),
                                   NumpyType(numpy.float64),
                                   )
                                  )


def name_trialfunction_gradient(element, restriction, component):
    rawname = "gradu" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)
    container = name_coefficientcontainer(restriction)

    # TODO
    #
    # This is just a temporary test used to create an A-matrix as
    # local operator class member. Right now it doesn't evaluate
    # anything.
    if get_option("sumfact") and restriction == Restriction.NONE:
        from dune.perftool.sumfact import evaluate_coefficient_gradient_sumfact
        evaluate_coefficient_gradient_sumfact(element, name, container, restriction, component)

    evaluate_coefficient_gradient(element, name, container, restriction, component)
    return name


def name_trialfunction(element, restriction, component):
    rawname = "u" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)
    container = name_coefficientcontainer(restriction)
    evaluate_coefficient(element, name, container, restriction, component)
    return name


def name_apply_function_gradient(element, restriction, component):
    rawname = "gradz_func" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)
    container = name_applycontainer(restriction)
    evaluate_coefficient_gradient(element, name, container, restriction, component)
    return name


def name_apply_function(element, restriction, component):
    rawname = "z_func" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)
    container = name_applycontainer(restriction)
    evaluate_coefficient(element, name, container, restriction, component)
    return name


def name_coefficientcontainer(restriction):
    name = restricted_name("x", restriction)
    return name


def name_applycontainer(restriction):
    name = restricted_name("z", restriction)
    return name


@pymbolic_expr
def pymbolic_coefficient(container, lfs, index):
    # TODO introduce a proper type for local function spaces!
    if isinstance(lfs, str):
        valuearg(lfs, dtype=NumpyType("str"))

    # If the LFS is not yet a pymbolic expression, make it one
    from pymbolic.primitives import Expression
    if not isinstance(lfs, Expression):
        lfs = Variable(lfs)

    return Call(CoefficientAccess(container), (lfs, Variable(index),))


def type_coefficientcontainer():
    return "X"


def name_jacobian(restriction1, restriction2):
    # Restrictions may only differ if NONE
    if (restriction1 == Restriction.NONE) or (restriction2 == Restriction.NONE):
        assert restriction1 == restriction2
    return restricted_name(restricted_name("jac", restriction1), restriction2)


def type_jacobian():
    return "J"


def name_residual(restriction):
    return restricted_name("r", restriction)


def type_residual():
    return "R"


def name_accumulation_variable(restrictions=(Restriction.NONE,)):
    ft = get_global_context_value("form_type")
    if ft == 'residual' or ft == 'jacobian_apply':
        return name_residual(*restrictions)
    if ft == 'jacobian':
        return name_jacobian(*restrictions)
    assert False


def type_accumulation_variable():
    ft = get_global_context_value("form_type")
    if ft == 'residual' or ft == 'jacobian_apply':
        return type_residual()
    if ft == 'jacobian':
        return type_jacobian()
    assert False
